/*********************************************************************************************************************************
 * Copyright (c) Des Herriott 1993 - 1994
 *               Erik Kunze   1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Des Herriott
 *          Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: emul.c,v 4.42 2000/11/26 18:39:49 erik Rel $";
#endif
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>
#include "cpu.h"
#include "resource.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "sbar.h"
#include "screen.h"
#include "ui.h"
#include "xkey.h"
#include "xwin.h"
#include "emul.h"
typedef struct _emulEvent {
	EmulEventFunc func;
	void *arg;
	int delay;
} emulEvent;
#define MAX_EVENTS						3
#define TIMER_INTERVAL					100
#define UPDATE_INTERVAL					2000
static void emulTimer(int);
static void emulScheduleEvents(void);
unsigned int EmulVideoLine;
unsigned int EmulVideoFrame;
static volatile int timerInterrupt;
static volatile int tstatesPerTimer;
static volatile int tstatesPerUpdate;
static emulEvent *eventTable;
void
EmulLoop(void)
{
	unsigned int cycles = 0;
	int tpu;
	int tpt;
	int ipu;
	int fps;
	int skipFrame;
	sigset_t smask;
	struct sigaction sigact;
	eventTable = Malloc(sizeof(emulEvent) * MAX_EVENTS, "EmulLoop");
	timerInterrupt = 0;
	EmulSetSpeed(CFGGET(speed));
	tpt = tstatesPerTimer;
	(void)sigfillset(&smask);
	(void)sigdelset(&smask, SIGALRM);
	sigact.sa_handler = emulTimer;
	(void)sigemptyset(&sigact.sa_mask);
#ifdef SA_RESTART
	sigact.sa_flags = SA_RESTART;
#else
	sigact.sa_flags = 0;
#endif
	(void)sigaction(SIGALRM, &sigact, (struct sigaction *)0);
	IntFrequency(TIMER_INTERVAL * 1000);
	ipu = UPDATE_INTERVAL / TIMER_INTERVAL;
	tpu = 0;
	fps = 0;
	for (;;)
	{
		if (EmulVideoLine == Machine->intLine)
		{
			Z80_INT(0xff, EmulVideoLine * Machine->tstatesPerLine);
		}
		cycles += Machine->tstatesPerLine;
		(void)Z80_Execute(cycles);
		skipFrame = (signed int)EmulVideoFrame % (CFGGET(fastmode) ? 10 : CFGGET(refresh));
		if (skipFrame == 0)
		{
			ScreenUpdateLine(EmulVideoLine);
		}
		if (++EmulVideoLine == Machine->tvLines)
		{
			EmulVideoLine = 0;
			EmulVideoFrame++;
			TSTATES -= cycles;
			cycles = 0;
#ifdef REAL_AUDIO
			AudioPlayBuffer();
#endif
			XwinProcessEvents();
			XkeyRefresh();
			if ((EmulVideoFrame % 25) == 0)
			{
				ScreenDoFlashing();
			}
			if (skipFrame == 0)
			{
				ScreenRefresh();
				fps++;
			}
		}
		emulScheduleEvents();
		if ((tpt -= Machine->tstatesPerLine) <= 0)
		{
			tpt += tstatesPerTimer;
			tpu += tstatesPerTimer;
			if (CFGGET(fastmode) == 0)
			{
				if (timerInterrupt == 0)
				{
					(void)sigsuspend(&smask);
				}
			}
			if ((ipu -= timerInterrupt) <= 0)
			{
				SbarSetSpeed(tpu / tstatesPerUpdate, fps / (UPDATE_INTERVAL / 1000));
				ipu = UPDATE_INTERVAL / TIMER_INTERVAL;
				tpu = 0;
				fps = 0;
			}
			timerInterrupt = 0;
		}
	}
}
static void
emulTimer(int signo)
{
	if (UiIsActive() == 0)
	{
		timerInterrupt++;
	}
}
void
EmulSetSpeed(int speed)
{
	double tpu;
	tstatesPerTimer = (int)Machine->tstatesPerFrame * speed / 2;
	tstatesPerTimer /= 1000 / TIMER_INTERVAL;
	tpu = (double)Machine->tstatesPerFrame * 50.0;
	tpu *= (double)(UPDATE_INTERVAL / 1000);
	tpu /= 100.0;
	tstatesPerUpdate = (unsigned int)tpu;
#ifdef REAL_AUDIO
	AudioControl(AUC_INIT);
#endif
}
EmulEventId
EmulAddEvent(EmulEventFunc func, void *arg, int delay)
{
	emulEvent *e;
	for (e = eventTable; e < &eventTable[MAX_EVENTS]; e++)
	{
		if (!e->func)
		{
			e->func = func;
			e->arg = arg;
			e->delay = delay;
			return (EmulEventId)(e - eventTable);
		}
	}
	Msg(M_FATAL, "event table overflow");
	return (EmulEventId)-1;
}
void
EmulRemoveEvent(EmulEventId slot, int call)
{
	emulEvent *e = &eventTable[slot];
	EmulEventFunc f;
	assert (slot >= 0 && slot < MAX_EVENTS);
	if (e->func)
	{
		f = e->func;
		e->func = (EmulEventFunc)0;
		if (call)
		{
			f(e->arg);
		}
	}
}
static void
emulScheduleEvents(void)
{
	emulEvent *e;
	EmulEventFunc f;
	for (e = eventTable; e < &eventTable[MAX_EVENTS]; e++)
	{
		if (e->func && --e->delay == 0)
		{
			f = e->func;
			e->func = (EmulEventFunc)0;
			f(e->arg);
		}
	}
}
