/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: plus3.c,v 1.20 2000/11/18 18:45:50 erik Rel $";
#endif
#include <assert.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#include "fdc.h"
#include "keyboard.h"
#include "kjoy.h"
#ifdef XZX_MF
#include "multiface.h"
#endif
#ifdef PRINTER
#include "printer.h"
#endif
#include "screen.h"
#include "tables.h"
#include "tape.h"
#include "plus3.h"
#define Y_INTERRUPT						0
#define Y_RETRACE						15
#define Y_TOP_BORDER					48
#define Y_SCREEN						192
#define Y_BOTTOM_BORDER					56
#define X_LEFT_BORDER					24
#define X_SCREEN						128
#define X_RIGHT							24
#define X_RETRACE						48
#define ULA_DRAWS_SCREEN				0x04
#define ULA_DRAWS_BORDER				0x20
#define ULA_HORIZONTAL_RETRACE			0x40
#define ULA_VERTICAL_RETRACE			0x80
#define ULA_DELAY_MASK					0x0f
#define T_STATES_OVERFLOW				23 + 19
#define IS_CONTENDED(addr)				(RPAGE((addr)) >= RAM4 && RPAGE((addr)) <= RAM7)
enum { ROM_UNKNOWN, ROM_PLUS3V40, ROM_PLUS3V41 };
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_IO) != 0) { x } }
#else
#define DEB(x)
#endif
static void plus3Reset(void);
static void plus3Nmi(void);
static uns8 plus3ReadOpcode(uns16);
static uns8 plus3ReadMem(uns16);
static void plus3WriteMem(uns16, uns8);
static void plus3InitPages(void);
static void plus3PageIn(int, int);
static inline void plus3OutBankm128(uns8);
static inline void plus3OutBank678(uns8);
static void plus3OutPort(uns16, uns8);
static uns8 plus3InPort(uns16);
void
Plus3Init(void)
{
	MachineSettings *msp;
	int checksum, romtype;
	unsigned int i, x, y;
	msp = &Machines[SP_3];
	msp->name = "Spectrum +3";
	msp->basicRom = SP3ROM3;
	msp->tvLines = Y_RETRACE + Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER;
	msp->tstatesPerLine = X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE;
	msp->tstatesPerFrame = msp->tvLines * msp->tstatesPerLine;
	msp->intLine = Y_INTERRUPT;
	msp->topBorder = Y_RETRACE + Y_TOP_BORDER;
	msp->ulaDelay = Malloc(msp->tstatesPerFrame + T_STATES_OVERFLOW, "Plus3Init");
	msp->reset = plus3Reset;
	msp->nmi = plus3Nmi;
	msp->readOpcode = plus3ReadOpcode;
	msp->readMem = plus3ReadMem;
	msp->writeMem = plus3WriteMem;
	msp->inPort = plus3InPort;
	msp->outPort = plus3OutPort;
	msp->initPages = plus3InitPages;
	msp->pageIn = plus3PageIn;
	if (!LoadRomImage(CFGGET(sp3Rom), SP3ROM0, 4))
	{
		checksum = GetRomChecksum(msp->basicRom, XZX_PAGESIZE);
		switch (checksum)
		{
			case 48934:
				romtype = ROM_PLUS3V40;
				break;
			case 10222:
				romtype = ROM_PLUS3V41;
				break;
			default:
				Msg(M_WARN, "unknown +3 ROM image - checksum %u", checksum);
				romtype = ROM_UNKNOWN;
				break;
		}
#ifndef NO_ED_TRAPS
		if (romtype != ROM_UNKNOWN)
		{
			RealMemory[msp->basicRom][0x0562] = 0xed;
			RealMemory[msp->basicRom][0x0563] = LOAD_BPT;
			RealMemory[msp->basicRom][0x04c6] = 0xed;
			RealMemory[msp->basicRom][0x04c7] = SAVE_BPT;
#ifdef XBELL_AUDIO
			RealMemory[msp->basicRom][0x03b5] = 0xed;
			RealMemory[msp->basicRom][0x03b6] = BEEP_BPT;
#endif
		}
#endif
	}
	for (i = 0, y = 0; y < Y_RETRACE; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_VERTICAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER + Y_SCREEN; y++)
	{
		for (x = 0; x < X_LEFT_BORDER; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_SCREEN;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	while (i < msp->tstatesPerFrame + T_STATES_OVERFLOW)
	{
		msp->ulaDelay[i++] = ULA_VERTICAL_RETRACE;
	}
	assert(i == msp->tstatesPerFrame + T_STATES_OVERFLOW);
}
static void
plus3Reset(void)
{
	Z80_Reset();
	Machine->initPages();
	Last0x7FFD = 0;
	ScreenSelect = RAM5;
	Last0x1FFD = 0;
#ifdef XZX_MF
	Multiface3Paged = 0;
#endif
	Last0xFFFD = 0;
	(void)memset(Psg, 0x00, sizeof(Psg));
#ifdef REAL_AUDIO
	AyReset();
#endif
	FdcReset();
}
static void
plus3Nmi(void)
{
	Z80_NMI((NmiCallbackFunc)0);
#ifdef XZX_MF
	(void)Machine->inPort(P_MF3IN);
#endif
}
static uns8
plus3ReadOpcode(uns16 addr)
{
	uns8 op = MEM(addr);
	if (IS_CONTENDED(addr))
	{
		TSTATES += (unsigned int)Machine->ulaDelay[TSTATES] & ULA_DELAY_MASK;
	}
	return op;
}
static uns8
plus3ReadMem(uns16 addr)
{
	uns8 byte = MEM(addr);
	if (IS_CONTENDED(addr))
	{
		TSTATES += (unsigned int)Machine->ulaDelay[TSTATES] & ULA_DELAY_MASK;
	}
	return byte;
}
static void
plus3WriteMem(uns16 addr, uns8 byte)
{
	if (!READ_ONLY(addr))
	{
		MEM(addr) = byte;
		if (RPAGE(addr) == ScreenSelect && (addr &= 0x3fff) < (uns16)(PIXEL_LENGTH + ATTR_LENGTH))
		{
			ScreenMarkDirty(addr);
		}
		if (IS_CONTENDED(addr))
		{
			TSTATES += (unsigned int)Machine->ulaDelay[TSTATES] & ULA_DELAY_MASK;
		}
	}
}
static void
plus3InitPages(void)
{
	Machine->pageIn(0, SP3ROM0);
	Machine->pageIn(1, RAM5);
	Machine->pageIn(2, RAM2);
	Machine->pageIn(3, RAM0);
}
static void
plus3PageIn(int z80page, int vpage)
{
	int i = z80page << 1;
	TheMemory[i] = RealMemory[vpage];
	RealPageNumbers[i] = vpage;
	i++;
	TheMemory[i] = RealMemory[vpage] + XZX_PAGESIZE / 2;
	switch (vpage)
	{
#ifdef XZX_MF
		case MF3ROM:
			RealPageNumbers[i] = MFRAM;
			break;
#endif
		default:
			RealPageNumbers[i] = vpage;
			break;
	}
}
static inline void
plus3OutBankm128(uns8 val)
{
	DEB(Msg(M_DEBUG,
			"bankm: RAM select %d, ROM select %d, screen %d, paging %sabled",
			val & B_SELRAM, (val & B_SELROM) >> 4, (val & B_SELSCREEN) >> 3,
			PAGING_DISABLED(Last0x7FFD) ? "dis" : "en"););
	if (Last0x7FFD == val || PAGING_DISABLED(Last0x7FFD))
	{
		return;
	}
	ScreenSelect = val & B_SELSCREEN ? RAM7 : RAM5;
	if (((val ^ Last0x7FFD) & B_SELSCREEN) != 0)
	{
		ScreenMarkDirtyAll();
	}
	Last0x7FFD = val;
	if (!EXTENDED_RAM(Last0x1FFD))
	{
		Machine->pageIn(1, RAM5);
		Machine->pageIn(2, RAM2);
		Machine->pageIn(3, RAM0 + (val & B_SELRAM));
#ifdef XZX_MF
		if (!Multiface3Paged)
#endif
		{
			Machine->pageIn(0, SP3ROM0 + (((val & B_SELROM) >> 4) | ((Last0x1FFD & B_SELROM3) >> 1)));
		}
		DEB(Msg(M_DEBUG, "bankm: horizontal ROM switch"););
	}
}
static inline void
plus3OutBank678(uns8 val)
{
	DEB(Msg(M_DEBUG, "bank678: %s", EXTENDED_RAM(Last0x1FFD) ? "extended RAM enabled" : "vertical ROM switch"););
	if (Last0x1FFD == val || PAGING_DISABLED(Last0x7FFD))
	{
		return;
	}
	if (!EXTENDED_RAM(val))
	{
#ifdef XZX_MF
		if (Multiface3Paged)
		{
			if (RPAGE(0) != MF3ROM)
			{
				Machine->pageIn(0, MF3ROM);
			}
		}
		else
#endif
		{
			Machine->pageIn(0, SP3ROM0 + (((Last0x7FFD & B_SELROM) >> 4) | ((val & B_SELROM3) >> 1)));
		}
	}
	else
	{
		switch ((val & B_SELRAM3) >> 1)
		{
			case 0:
				Machine->pageIn(0, RAM0);
				Machine->pageIn(1, RAM1);
				Machine->pageIn(2, RAM2);
				Machine->pageIn(3, RAM3);
				DEB(Msg(M_DEBUG, "bank678: 0,1,2,3"););
				break;
			case 1:
				Machine->pageIn(0, RAM4);
				Machine->pageIn(1, RAM5);
				Machine->pageIn(2, RAM6);
				Machine->pageIn(3, RAM7);
				DEB(Msg(M_DEBUG, "bank678: 4,5,6,7"););
				break;
			case 2:
				Machine->pageIn(0, RAM4);
				Machine->pageIn(1, RAM5);
				Machine->pageIn(2, RAM6);
				Machine->pageIn(3, RAM3);
				DEB(Msg(M_DEBUG, "bank678: 4,5,6,3"););
				break;
			case 3:
				Machine->pageIn(0, RAM4);
				Machine->pageIn(1, RAM7);
				Machine->pageIn(2, RAM6);
				Machine->pageIn(3, RAM3);
				DEB(Msg(M_DEBUG, "bank678: 4,7,6,3"););
				break;
		}
	}
#ifdef PLUS3_PRINTER
	if ((val ^ Last0x1FFD) & B_STROBE)
	{
		Plus3CentronicsStrobe(val);
	}
#endif
	Last0x1FFD = val;
}
static void
plus3OutPort(uns16 port, uns8 val)
{
	DEB(Msg(M_DEBUG, "writing <%02x> to port <%04x>", val, port););
	if ((port & B_ULA) == (P_ULA & B_ULA))
	{
		uns8 changes = val ^ Last0xFE;
		if ((changes & B_BORDER) != 0)
		{
			ScreenSetBorder(val & B_BORDER);
		}
#ifdef REAL_AUDIO
		if ((changes & (B_SPEAKER | B_MIC)) != 0)
		{
			SpeakerOutByte(val);
		}
#endif
		Last0xFE = val;
	}
	if ((port & B_128) == (P_128 & B_128))
	{
		switch (port & 0xc000)
		{
			case P_SNDCONTROL & 0xc000:
				Last0xFFFD = val;
				break;
			case P_SNDDATA & 0xc000:
				if (Last0xFFFD < NELEM(Psg))
				{
					Psg[Last0xFFFD] = val;
					if (Last0xFFFD == 14)
					{
						DEB(Msg(M_DEBUG, "writing AY register 14"););
					}
#ifdef REAL_AUDIO
					else if (Last0xFFFD < 14)
					{
						AyOutByte(Last0xFFFD, val);
					}
#endif
				}
				else
				{
					DEB(Msg(M_DEBUG, "writing AY register %d", Last0xFFFD););
				}
				break;
			case P_BANK128 & 0xc000:
				plus3OutBankm128(val);
				break;
			default:
				switch (port & 0x3000)
				{
#ifdef PLUS3_PRINTER
					case P_CENTRONICS & 0x3000:
						Plus3CentronicsOutByte(val);
						break;
#endif
					case P_BANK3 & 0x3000:
						plus3OutBank678(val);
						break;
					case P_FDCCONTROL & 0x3000:
						break;
					case P_FDCDATA & 0x3000:
						FdcDataOutByte(val);
						break;
				}
		}
	}
}
static uns8
plus3InPort(uns16 port)
{
	uns8 res;
	INBYTE_KJOY(port, res, quit);
	if ((port & B_ULA) == (P_ULA & B_ULA))
	{
		res = 0xbf;
		INBYTE_KEYBOARD(port, res);
		res ^= TapeInGetBit();
		goto quit;
	}
	if ((port & B_128) == (P_128 & B_128))
	{
		switch (port & 0xc000)
		{
			case P_SNDCONTROL & 0xc000:
				if (Last0xFFFD == 14)
				{
					DEB(Msg(M_DEBUG, "reading AY register 14"););
					res = Psg[7] & 0x40 ? (Psg[14] & 0x1f) | 0x20 : 0x3f;
				}
				else if (Last0xFFFD < 14)
				{
					res = (Psg[Last0xFFFD] & PsgMask[Last0xFFFD]);
				}
				else if (Last0xFFFD == 15)
				{
					res = Psg[7] & 0x80 ? Psg[15] : 0xff;
				}
				else
				{
					res = 0x00;
				}
				goto quit;
			case P_SNDDATA & 0xc000:
				res = 0xff;
				goto quit;
			case P_BANK128 & 0xc000:
				break;
			default:
				switch (port & 0x3000)
				{
#ifdef PLUS3_PRINTER
					case P_CENTRONICS & 0x3000:
						res = Plus3CentronicsInByte();
						goto quit;
#endif
					case P_BANK3 & 0x3000:
						break;
					case P_FDCCONTROL & 0x3000:
						res = FdcControlInByte();
						goto quit;
					case P_FDCDATA & 0x3000:
						res = FdcDataInByte();
						goto quit;
				}
		}
	}
#ifdef XZX_MF
	INBYTE_MF3(port, res, quit);
#endif
	res = 0xff;
  quit:
	DEB(Msg(M_DEBUG, "reading <%02x> from port <%04x>", res, port););
	return res;
}
