/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef XZX_SCORPION
#ifndef lint
static char rcsid[] = "$Id: scorpion.c,v 1.23 2000/11/26 18:56:33 erik Rel $";
#endif
#include <assert.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#include "keyboard.h"
#include "kjoy.h"
#ifdef XZX_KMOUSE
#include "kmouse.h"
#endif
#include "screen.h"
#include "tables.h"
#include "tape.h"
#include "trdos.h"
#include "scorpion.h"
#define Y_INTERRUPT						0
#define Y_TOP_BORDER					64
#define Y_SCREEN						192
#define Y_BOTTOM_BORDER					40
#define Y_RETRACE						16
#define X_LEFT_BORDER					24
#define X_SCREEN						128
#define X_RIGHT							32
#define X_RETRACE						40
#define ULA_DRAWS_SCREEN				0x00
#define ULA_DRAWS_BORDER				0x20
#define ULA_HORIZONTAL_RETRACE			0x40
#define ULA_VERTICAL_RETRACE			0x80
#define P_BANKX							0x1ffd
#define B_BANKX							0xe002
#define B_SPECIALX						0x01
#define B_SELROMX						0x02
#define B_SELRAMX						0x10
#define EXTENDED_MEM(port)				((port) & (B_SPECIALX | B_SELROMX))
enum { ROM_UNKNOWN, ROM_SCORPION };
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_IO) != 0) { x } }
#else
#define DEB(x)
#endif
static void scorpionReset(void);
static void scorpionNmi(void);
static uns8 scorpionReadOpcode(uns16);
static uns8 scorpionReadMem(uns16);
static void scorpionWriteMem(uns16, uns8);
static void scorpionInitPages(void);
static void scorpionPageIn(int, int);
static inline void scorpionOutBankm128(uns8);
static inline void scorpionOutBankmX(uns8);
static void scorpionOutPort(uns16, uns8);
static uns8 scorpionInPort(uns16);
void
ScorpionInit(void)
{
	MachineSettings *msp;
	int checksum, romtype;
	unsigned int i, x, y;
	msp = &Machines[SP_SCORPION];
	msp->name = "Scorpion 256";
	msp->basicRom = SCORROM1;
	msp->tvLines = Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER + Y_RETRACE;
	msp->tstatesPerLine = X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE;
	msp->tstatesPerFrame = msp->tvLines * msp->tstatesPerLine;
	msp->intLine = Y_INTERRUPT;
	msp->topBorder = Y_TOP_BORDER;
	msp->ulaDelay = Malloc(msp->tstatesPerFrame, "ScorpionInit");
	msp->reset = scorpionReset;
	msp->nmi = scorpionNmi;
	msp->readOpcode = scorpionReadOpcode;
	msp->readMem = scorpionReadMem;
	msp->writeMem = scorpionWriteMem;
	msp->inPort = scorpionInPort;
	msp->outPort = scorpionOutPort;
	msp->initPages = scorpionInitPages;
	msp->pageIn = scorpionPageIn;
	if (!LoadRomImage(CFGGET(scorpionRom), SCORROM0, 4))
	{
		checksum = GetRomChecksum(msp->basicRom, XZX_PAGESIZE);
		switch (checksum)
		{
			case 22821:
				romtype = ROM_SCORPION;
				break;
			default:
				Msg(M_WARN, "unknown Scorpion ROM image - checksum %u",
					checksum);
				romtype = ROM_UNKNOWN;
				break;
		}
#ifndef NO_ED_TRAPS
		if (romtype != ROM_UNKNOWN)
		{
			RealMemory[msp->basicRom][0x0562] = 0xed;
			RealMemory[msp->basicRom][0x0563] = LOAD_BPT;
			RealMemory[msp->basicRom][0x04c6] = 0xed;
			RealMemory[msp->basicRom][0x04c7] = SAVE_BPT;
#ifdef XBELL_AUDIO
			RealMemory[msp->basicRom][0x03b5] = 0xed;
			RealMemory[msp->basicRom][0x03b6] = BEEP_BPT;
#endif
		}
#endif
	}
	for (i = 0, y = 0; y < Y_TOP_BORDER; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_TOP_BORDER + Y_SCREEN; y++)
	{
		for (x = 0; x < X_LEFT_BORDER; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_SCREEN;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER + Y_RETRACE; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_VERTICAL_RETRACE;
		}
	}
	assert(i == msp->tstatesPerFrame);
}
static void
scorpionReset(void)
{
	Z80_Reset();
	Machine->initPages();
	Last0x7FFD = 0;
	ScreenSelect = RAM5;
	Last0x1FFD = 0;
	Last0xFFFD = 0;
	(void)memset(Psg, 0x00, sizeof(Psg));
#ifdef REAL_AUDIO
	AyReset();
#endif
}
static void
scorpionNmi(void)
{
	if (PC >= 0x4000)
	{
		Z80_NMI((NmiCallbackFunc)0);
		(void)Machine->pageIn(0, SCORROM2);
	}
}
static uns8
scorpionReadOpcode(uns16 addr)
{
	if (addr >= 0x4000)
	{
		if (RPAGE(0) == SCORROM3)
		{
			Machine->pageIn(0, Machine->basicRom);
		}
	}
	else if ((addr & 0xff00) == 0x3d00)
	{
		if (RPAGE(0) == Machine->basicRom)
		{
			Machine->pageIn(0, SCORROM3);
		}
	}
	return MEM(addr);
}
static uns8
scorpionReadMem(uns16 addr)
{
	return MEM(addr);
}
static void
scorpionWriteMem(uns16 addr, uns8 byte)
{
	if (!READ_ONLY(addr))
	{
		MEM(addr) = byte;
		if (RPAGE(addr) == ScreenSelect && (addr &= 0x3fff) < (uns16)(PIXEL_LENGTH + ATTR_LENGTH))
		{
			ScreenMarkDirty(addr);
		}
	}
}
static void
scorpionInitPages(void)
{
	Machine->pageIn(0, SCORROM0);
	Machine->pageIn(1, RAM5);
	Machine->pageIn(2, RAM2);
	Machine->pageIn(3, RAM0);
}
static void
scorpionPageIn(int z80page, int vpage)
{
	int i = z80page << 1;
	TheMemory[i] = RealMemory[vpage];
	RealPageNumbers[i] = vpage;
	i++;
	TheMemory[i] = RealMemory[vpage] + XZX_PAGESIZE / 2;
	RealPageNumbers[i] = vpage;
}
static inline void
scorpionOutBankm128(uns8 val)
{
	DEB(Msg(M_DEBUG,
			"bankm: RAM select %x, ROM select %d, screen %d, paging %sabled",
			((Last0x1FFD & B_SELRAMX) >> 1) | (val & B_SELRAM),
			(val & B_SELROM) >> 4, (val & B_SELSCREEN) >> 3,
			PAGING_DISABLED(Last0x7FFD) ? "dis" : "en"););
	if (PAGING_DISABLED(Last0x7FFD))
	{
		return;
	}
	ScreenSelect = val & B_SELSCREEN ? RAM7 : RAM5;
	if (((val  ^ Last0x7FFD) & B_SELSCREEN) != 0)
	{
		ScreenMarkDirtyAll();
	}
	Last0x7FFD = val;
	Machine->pageIn(3, RAM0 + (((Last0x1FFD & B_SELRAMX) >> 1) | (val & B_SELRAM)));
	if (!EXTENDED_MEM(Last0x1FFD))
	{
		Machine->pageIn(0, SCORROM0 + ((val & B_SELROM) >> 4));
	}
}
static inline void
scorpionOutBankmX(uns8 val)
{
	int page0;
	DEB(Msg(M_DEBUG,
			"bankx: RAM select %x, RAM8 %d, ROM2 %d",
			((val & B_SELRAMX) >> 1) | (Last0x7FFD & B_SELRAM),
			val & B_SPECIALX, (val & B_SELROMX) >> 1););
	switch (EXTENDED_MEM(val))
	{
		default:
			page0 = SCORROM0 + ((Last0x7FFD & B_SELROM) >> 4);
			break;
		case B_SPECIALX:
			page0 = RAM8;
			break;
		case B_SPECIALX | B_SELROMX:
			Msg(M_WARN, "RAM and ROM together at page 0 is impossible");
		case B_SELROMX:
			page0 = SCORROM2;
			break;
	}
	Machine->pageIn(0, page0);
	Machine->pageIn(3, RAM0 + (((val & B_SELRAMX) >> 1) | (Last0x7FFD & B_SELRAM)));
	Last0x1FFD = val;
}
static void
scorpionOutPort(uns16 port, uns8 val)
{
	DEB(Msg(M_DEBUG, "writing <%02x> to port <%04x>", val, port););
	if (RPAGE(0) >= SCORROM2)
	{
		TrdosOutByte(port, val);
	}
	if ((port & B_ULA) == (P_ULA & B_ULA))
	{
		uns8 changes = val ^ Last0xFE;
		if ((changes & B_BORDER) != 0)
		{
			ScreenSetBorder(val & B_BORDER);
		}
#ifdef REAL_AUDIO
		if ((changes & (B_SPEAKER | B_MIC)) != 0)
		{
			SpeakerOutByte(val);
		}
#endif
		Last0xFE = val;
	}
	if ((port & B_128) == (P_128 & B_128))
	{
		switch (port & 0xc000)
		{
			case P_SNDCONTROL & 0xc000:
				Last0xFFFD = val;
				break;
			case P_SNDDATA & 0xc000:
				if (Last0xFFFD < NELEM(Psg))
				{
					Psg[Last0xFFFD] = val;
					if (Last0xFFFD == 14)
					{
						DEB(Msg(M_DEBUG, "writing AY register 14"););
					}
#ifdef REAL_AUDIO
					else if (Last0xFFFD < 14)
					{
						AyOutByte(Last0xFFFD, val);
					}
#endif
				}
				else
				{
					DEB(Msg(M_DEBUG, "writing AY register %d", Last0xFFFD););
				}
				break;
			case P_BANK128 & 0xc000:
				scorpionOutBankm128(val);
				break;
			default:
				switch (port & 0x3000)
				{
					case P_BANKX & 0x3000:
						scorpionOutBankmX(val);
						break;
				}
		}
	}
}
static uns8
scorpionInPort(uns16 port)
{
	uns8 res;
	if (RPAGE(0) >= SCORROM2
		&& (port & 0x001f) == 0x001f
		&& ((port & 0x0080) == 0x0000 || (port & 0x00f0) == 0x00f0))
	{
		res = TrdosInByte(port);
		goto quit;
	}
#ifdef XZX_KMOUSE
	INBYTE_KMOUSE(port, res, quit);
#endif
	INBYTE_KJOY(port, res, quit);
	if ((port & B_ULA) == (P_ULA & B_ULA))
	{
		res = 0x1f;
		INBYTE_KEYBOARD(port, res);
		res ^= TapeInGetBit();
		goto quit;
	}
	if ((port & B_128) == (P_128 & B_128))
	{
		switch (port & 0xc000)
		{
			case P_SNDCONTROL & 0xc000:
				if (Last0xFFFD == 14)
				{
					DEB(Msg(M_DEBUG, "reading AY register 14"););
					res = Psg[7] & 0x40 ? (Psg[14] & 0x1f) | 0x20 : 0x3f;
				}
				else if (Last0xFFFD < 14)
				{
					res = (Psg[Last0xFFFD] & PsgMask[Last0xFFFD]);
				}
				else if (Last0xFFFD == 15)
				{
					res = (Psg[7] & 0x80 ? Psg[15] : 0xff);
				}
				else
				{
					res = 0xff;
				}
				goto quit;
			case P_SNDDATA & 0xc000:
				res = 0xff;
				goto quit;
			case P_BANK128 & 0xc000:
				Msg(M_WARN, "turbo mode on - not supported by XZX");
				res = 0xff;
				goto quit;
			case P_BANKX & 0xc000:
				Msg(M_WARN, "turbo mode off");
				res = 0xff;
				goto quit;
		}
	}
	if (Machine->ulaDelay[TSTATES % Machine->tstatesPerFrame] == ULA_DRAWS_SCREEN)
	{
		unsigned int tstatesInLine = TSTATES % Machine->tstatesPerLine;
		unsigned int line = EmulVideoLine - Machine->topBorder;
		res = RealMemory[ScreenSelect][(tstatesInLine & 0x02 ? Line2Attr[line] : Line2Pix[line]) + (tstatesInLine >> 2)];
	}
	else
	{
		res = 0xff;
	}
  quit:
	DEB(Msg(M_DEBUG, "reading <%02x> from port <%04x>", res, port););
	return res;
}
#endif
