/*********************************************************************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the copyright holder not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 ********************************************************************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: spec128.c,v 1.20 2000/11/26 18:55:32 erik Rel $";
#endif
#include <assert.h>
#include "cpu.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#ifdef XZX_IF1
#include "if1.h"
#endif
#include "keyboard.h"
#include "kjoy.h"
#ifdef XZX_MF
#include "multiface.h"
#endif
#ifdef XZX_MGT
#include "mgt.h"
#endif
#ifdef PRINTER
#include "printer.h"
#endif
#include "screen.h"
#include "tables.h"
#include "tape.h"
#ifdef XZX_TRDOS
#include "trdos.h"
#endif
#include "spec128.h"
#define Y_INTERRUPT						0
#define Y_RETRACE						15
#define Y_TOP_BORDER					48
#define Y_SCREEN						192
#define Y_BOTTOM_BORDER					56
#define X_LEFT_BORDER					26
#define X_SCREEN						128
#define X_RIGHT							26
#define X_RETRACE						48
#define ULA_DRAWS_SCREEN				0x04
#define ULA_DRAWS_BORDER				0x20
#define ULA_HORIZONTAL_RETRACE			0x40
#define ULA_VERTICAL_RETRACE			0x80
#define ULA_DELAY_MASK					0x0f
#define T_STATES_OVERFLOW				23 + 19
#define IS_CONTENDED(addr)				(RPAGE((addr)) >= RAM4 && RPAGE((addr)) <= RAM7)
enum { ROM_UNKNOWN, ROM_128, ROM_PLUS2 };
#ifdef DEBUG
#define DEB(x)							{ if ((CFGGET(debug) & D_IO) != 0) { x } }
#else
#define DEB(x)
#endif
static void spec128Reset(void);
static void spec128Nmi(void);
static uns8 spec128ReadOpcode(uns16);
static uns8 spec128ReadMem(uns16);
static void spec128WriteMem(uns16, uns8);
static void spec128InitPages(void);
static void spec128PageIn(int, int);
static inline void spec128OutBankm128(uns8);
static void spec128OutPort(uns16, uns8);
static uns8 spec128InPort(uns16);
void
Spec128Init(void)
{
	MachineSettings *msp;
	int checksum, romtype;
	unsigned i, x, y;
	msp = &Machines[SP_128];
	msp->name = "Spectrum 128K";
	msp->basicRom = SP128ROM1;
	msp->tvLines = Y_RETRACE + Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER;
	msp->tstatesPerLine = X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE;
	msp->tstatesPerFrame = msp->tvLines * msp->tstatesPerLine;
	msp->intLine = Y_INTERRUPT;
	msp->topBorder = Y_RETRACE + Y_TOP_BORDER;
	msp->ulaDelay = Malloc(msp->tstatesPerFrame + T_STATES_OVERFLOW, "Spec128Init");
	msp->reset = spec128Reset;
	msp->nmi = spec128Nmi;
	msp->readOpcode = spec128ReadOpcode;
	msp->readMem = spec128ReadMem;
	msp->writeMem = spec128WriteMem;
	msp->inPort = spec128InPort;
	msp->outPort = spec128OutPort;
	msp->initPages = spec128InitPages;
	msp->pageIn = spec128PageIn;
	if (!LoadRomImage(CFGGET(sp128Rom), SP128ROM0, 2))
	{
		checksum = GetRomChecksum(msp->basicRom, XZX_PAGESIZE);
		switch (checksum)
		{
			case 44809:
				romtype = ROM_128;
				break;
			case 32991:
				msp->name = "Spectrum +2";
				romtype = ROM_PLUS2;
				break;
			default:
				Msg(M_WARN, "unknown 128K ROM image - checksum %u", checksum);
				romtype = ROM_UNKNOWN;
				break;
		}
#ifndef NO_ED_TRAPS
		if (romtype != ROM_UNKNOWN)
		{
			RealMemory[msp->basicRom][0x0562] = 0xed;
			RealMemory[msp->basicRom][0x0563] = LOAD_BPT;
			RealMemory[msp->basicRom][0x04c6] = 0xed;
			RealMemory[msp->basicRom][0x04c7] = SAVE_BPT;
#ifdef XBELL_AUDIO
			RealMemory[msp->basicRom][0x03b5] = 0xed;
			RealMemory[msp->basicRom][0x03b6] = BEEP_BPT;
#endif
		}
#endif
	}
	for (i = 0, y = 0; y < Y_RETRACE; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_VERTICAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER + Y_SCREEN; y++)
	{
		for (x = 0; x < X_LEFT_BORDER; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_SCREEN;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	for (; y < Y_RETRACE + Y_TOP_BORDER + Y_SCREEN + Y_BOTTOM_BORDER; y++)
	{
		for (x = 0; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT; x++)
		{
			msp->ulaDelay[i++] = ULA_DRAWS_BORDER;
		}
		for (; x < X_LEFT_BORDER + X_SCREEN + X_RIGHT + X_RETRACE; x++)
		{
			msp->ulaDelay[i++] = ULA_HORIZONTAL_RETRACE;
		}
	}
	while (i < msp->tstatesPerFrame + T_STATES_OVERFLOW)
	{
		msp->ulaDelay[i++] = ULA_VERTICAL_RETRACE;
	}
	assert(i == msp->tstatesPerFrame + T_STATES_OVERFLOW);
}
static void
spec128Reset(void)
{
	Z80_Reset();
	Machine->initPages();
	Last0x7FFD = 0;
	ScreenSelect = RAM5;
	Last0xFFFD = 0;
	(void)memset(Psg, 0x00, sizeof(Psg));
#ifdef REAL_AUDIO
	AyReset();
#endif
#ifdef XZX_TRDOS
	TrdosReset();
#endif
#ifdef XZX_MGT
	MgtReset();
#endif
}
static void
spec128Nmi(void)
{
	Z80_NMI((NmiCallbackFunc)0);
#ifdef XZX_MF
	(void)Machine->inPort(P_MF128IN);
#endif
#ifdef XZX_TRDOS
	if (CFGGET(trdos) > 0)
	{
		Machine->pageIn(0, TRDOSROM);
	}
#endif
#ifdef XZX_MGT
	if (CFGGET(mgt) > 0)
	{
		Machine->pageIn(0, MGTROM);
	}
#endif
}
static uns8
spec128ReadOpcode(uns16 addr)
{
	uns8 op;
#ifdef XZX_IF1
	CHECK_AND_PAGEIN_IF1(addr);
#endif
#ifdef XZX_TRDOS
	CHECK_TRDOS(addr);
#endif
#ifdef XZX_MGT
	CHECK_MGT(addr);
#endif
	op = MEM(addr);
	if (IS_CONTENDED(addr))
	{
		TSTATES += (unsigned int)Machine->ulaDelay[TSTATES] & ULA_DELAY_MASK;
	}
#ifdef XZX_IF1
	CHECK_AND_PAGEOUT_IF1(addr);
#endif
	return op;
}
static uns8
spec128ReadMem(uns16 addr)
{
	uns8 byte = MEM(addr);
	if (IS_CONTENDED(addr))
	{
		TSTATES += (unsigned int)Machine->ulaDelay[TSTATES] & ULA_DELAY_MASK;
	}
	return byte;
}
static void
spec128WriteMem(uns16 addr, uns8 byte)
{
	if (!READ_ONLY(addr))
	{
		MEM(addr) = byte;
		if (RPAGE(addr) == RAM5 && (addr &= 0x3fff) < (uns16)(PIXEL_LENGTH + ATTR_LENGTH))
		{
			ScreenMarkDirty(addr);
		}
		if (IS_CONTENDED(addr))
		{
			TSTATES += (unsigned int)Machine->ulaDelay[TSTATES] & ULA_DELAY_MASK;
		}
	}
}
static void
spec128InitPages(void)
{
	Machine->pageIn(0, SP128ROM0);
	Machine->pageIn(1, RAM5);
	Machine->pageIn(2, RAM2);
	Machine->pageIn(3, RAM0);
}
static void
spec128PageIn(int z80page, int vpage)
{
	int i = z80page << 1;
	TheMemory[i] = RealMemory[vpage];
	RealPageNumbers[i] = vpage;
	i++;
	TheMemory[i] = RealMemory[vpage] + XZX_PAGESIZE / 2;
	switch (vpage)
	{
#ifdef XZX_MF
		case MF128ROM:
			RealPageNumbers[i] = MFRAM;
			break;
#endif
#ifdef XZX_MGT
		case MGTROM:
			RealPageNumbers[i] = MGTRAM;
			break;
#endif
		default:
			RealPageNumbers[i] = vpage;
			break;
	}
}
static inline void
spec128OutBankm128(uns8 val)
{
	DEB(Msg(M_DEBUG,
			"bankm: RAM select %d, ROM select %d, screen %d, paging %sabled",
			val & B_SELRAM, (val & B_SELROM) >> 4, (val & B_SELSCREEN) >> 3,
			PAGING_DISABLED(Last0x7FFD) ? "dis" : "en"););
	if (Last0x7FFD == val || PAGING_DISABLED(Last0x7FFD))
	{
		return;
	}
	ScreenSelect = val & B_SELSCREEN ? RAM7 : RAM5;
	if (((val ^ Last0x7FFD) & B_SELSCREEN) != 0)
	{
		ScreenMarkDirtyAll();
	}
	Last0x7FFD = val;
	Machine->pageIn(3, RAM0 + (val & B_SELRAM));
#ifdef XZX_MF
	if (RPAGE(0) != MF128ROM)
#endif
	{
		Machine->pageIn(0, SP128ROM0 + ((val & B_SELROM) >> 4));
	}
}
static void
spec128OutPort(uns16 port, uns8 val)
{
	DEB(Msg(M_DEBUG, "writing <%02x> to port <%04x>", val, port););
#ifdef XZX_TRDOS
	OUTBYTE_TRDOS(port, val);
#endif
#ifdef XZX_MGT
	OUTBYTE_MGT(port, val);
#endif
	if ((port & B_ULA) == (P_ULA & B_ULA))
	{
		uns8 changes = val ^ Last0xFE;
		if ((changes & B_BORDER) != 0)
		{
			ScreenSetBorder(val & B_BORDER);
		}
#ifdef REAL_AUDIO
		if ((changes & (B_SPEAKER | B_MIC)) != 0)
		{
			SpeakerOutByte(val);
		}
#endif
		Last0xFE = val;
		TSTATES += (unsigned int)Machine->ulaDelay[TSTATES] & ULA_DELAY_MASK;
	}
	if ((port & B_128) == (P_128 & B_128))
	{
		switch (port & 0xc000)
		{
			case P_SNDCONTROL & 0xc000:
				Last0xFFFD = val;
				break;
			case P_SNDDATA & 0xc000:
				if (Last0xFFFD < NELEM(Psg))
				{
					Psg[Last0xFFFD] = val;
					if (Last0xFFFD == 14)
					{
						DEB(Msg(M_DEBUG, "writing AY register 14"););
					}
#ifdef REAL_AUDIO
					else if (Last0xFFFD < 14)
					{
						AyOutByte(Last0xFFFD, val);
					}
#endif
				}
				else
				{
					DEB(Msg(M_DEBUG, "writing AY register %d", Last0xFFFD););
				}
				break;
			case P_BANK128 & 0xc000:
				spec128OutBankm128(val);
				break;
		}
	}
#ifdef XZX_IF1
	OUTBYTE_IF1(port, val);
#endif
#ifdef PSEUDO_IO
	if (port == PSEUDO_STDOUT)
	{
		(void)putc((int)val, stdout);
	}
	else if (port == PSEUDO_STDERR)
	{
		(void)putc((int)val, stderr);
	}
#endif
}
static uns8
spec128InPort(uns16 port)
{
	uns8 res;
#ifdef XZX_TRDOS
	INBYTE_TRDOS(port, res, quit);
#endif
#ifdef XZX_MGT
	INBYTE_MGT(port, res, quit);
#endif
	INBYTE_KJOY(port, res, quit);
	if ((port & B_ULA) == (P_ULA & B_ULA))
	{
		res = Last0xFE & B_SPEAKER ? 0xff : 0xbf;
		INBYTE_KEYBOARD(port, res);
		res ^= TapeInGetBit();
		TSTATES += 1U;
		goto quit;
	}
	if ((port & B_128) == (P_128 & B_128))
	{
		switch (port & 0xc000)
		{
			case P_SNDCONTROL & 0xc000:
				if (Last0xFFFD == 14)
				{
					DEB(Msg(M_DEBUG, "reading AY register 14"););
					res = Psg[7] & 0x40 ? (Psg[14] & 0x1f) | 0x20 : 0x3f;
				}
				else if (Last0xFFFD < 14)
				{
					res = (Psg[Last0xFFFD] & PsgMask[Last0xFFFD]);
				}
				else if (Last0xFFFD == 15)
				{
					res = Psg[7] & 0x80 ? Psg[15] : 0xff;
				}
				else
				{
					res = 0x00;
				}
				goto quit;
			case P_SNDDATA & 0xc000:
				res = 0xff;
				goto quit;
			case P_BANK128 & 0xc000:
				break;
		}
	}
#ifdef XZX_IF1
	INBYTE_IF1(port, res, quit);
#endif
#ifdef XZX_MF
	INBYTE_MF128(port, res, quit);
#endif
#ifdef PSEUDO_IO
	if (port == PSEUDO_STDIN)
	{
		res = (uns8)getchar();
		goto quit;
	}
#endif
	if ((Machine->ulaDelay[TSTATES % Machine->tstatesPerFrame] & ULA_DELAY_MASK) == ULA_DRAWS_SCREEN)
	{
		unsigned int tstatesInLine = TSTATES % Machine->tstatesPerLine;
		unsigned int line = EmulVideoLine - Machine->topBorder;
		res = RealMemory[ScreenSelect][(tstatesInLine & 0x02 ? Line2Attr[line] : Line2Pix[line]) + (tstatesInLine >> 2)];
	}
	else
	{
		res = 0xff;
	}
  quit:
	DEB(Msg(M_DEBUG, "reading <%02x> from port <%04x>", res, port););
	return res;
}
