/********************************************************************************
 * Copyright (c) Erik Kunze 1995 - 2000
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 *******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef REAL_AUDIO
#ifndef lint
static char rcsid[] = "$Id: audio.c,v 1.18 2000/11/26 18:51:33 erik Rel $";
#endif
#include <stdlib.h>
#ifdef HAVE_MEMORY_H
# include <memory.h>
#endif
#include <assert.h>
#include "debug.h"
#include "resource.h"
#include "machine.h"
#include "util.h"
#include "audio.h"
#ifdef DEBUG
#define DEB(x)			{ if (CFGGET(debug) & D_AUDIO) { x } }
#else
#define DEB(x)
#endif
static int audioOpen(void);
static void audioClose(void);
static void audioDummyOutByte(void);
AudioSettings Audio;
static unsigned char *mixerBuffer;
static size_t mixerSize;
void
AudioInit(void)
{
	SpeakerOutByte = (void (*)(uns8))audioDummyOutByte;
	AyOutByte = (void (*)(uns8, uns8))audioDummyOutByte;
	if (audioOpen() < 0)
	{
		Msg(M_WARN, "audio initialisation failed");
		CFGSET(audio, -1);
		return;
	}
	OnQuit(audioClose);
	AudioControl(AUC_INIT);
	AudioSetVolume(AUV_SP, CFGGET(audio) > 0 ? 2U : 0U);
	AudioSetVolume(AUV_AY, CFGGET(audio) > 0 ? 4U : 0U);
}
static int
audioOpen(void)
{
	int sample, stereo;
	if (Audio.active)
	{
		return 1;
	}
	sample = CFGGET(audioSample);
	stereo = 0;
	if (OsdOpenAudioDevice(CFGGET(audioDev), &sample, &stereo) < 0)
	{
		return -1;
	}
	CFGSET(audioSample, sample);
#ifdef REGISTERED
	if (!stereo)
	{
		CFGSET(audioStereo, AY_MONO);
	}
#endif
	Audio.sampleRate = sample;
	Audio.stereo = stereo;
	DEB(Msg(M_DEBUG, "AU: sampleRate       %u", sample);
		Msg(M_DEBUG, "AU: stereo           %u", stereo););
	Audio.active = 1;
	return 0;
}
static void
audioClose(void)
{
	if (!Audio.active)
	{
		return;
	}
	(void)OsdDrainAudioDevice();
	OsdCloseAudioDevice();
	Audio.active = 0;
}
void
AudioControl(int cmd)
{
	if (!Audio.active)
	{
		return;
	}
	switch (cmd)
	{
		case AUC_INIT:
			Audio.samplesPerFrame = Audio.sampleRate * 2 / CFGGET(speed);
			Audio.tstatesPerSample = Machine->tstatesPerFrame /
									 Audio.samplesPerFrame;
			DEB(Msg(M_DEBUG, "AU: tstatesPerSample %u", Audio.tstatesPerSample);
				Msg(M_DEBUG, "AU: samplesPerFrame  %u", Audio.samplesPerFrame););
			SpeakerInit();
			AyInit();
			mixerSize = Audio.samplesPerFrame << Audio.stereo;
			if (mixerBuffer)
			{
				free(mixerBuffer);
			}
			mixerBuffer = Malloc(mixerSize, "AudioControl");
			break;
		case AUC_SUSPEND:
			(void)OsdDrainAudioDevice();
			break;
		case AUC_RESUME:
			break;
		case AUC_ON:
			if (Audio.spVol)
			{
				SpeakerOutByte = SpeakerOut;
			}
			if (Audio.ayVol)
			{
				AyOutByte = AyOut;
			}
			break;
		case AUC_OFF:
			SpeakerOutByte = (void (*)(uns8))audioDummyOutByte;
			AyOutByte = (void (*)(uns8, uns8))audioDummyOutByte;
			break;
	}
}
void
AudioSetVolume(int channel, unsigned int volume)
{
	switch (channel)
	{
		case AUV_SP:
			Audio.spVol = volume;
			break;
		case AUV_AY:
			Audio.ayVol = volume;
			break;
	}
	if (!Audio.active || CFGGET(fastmode))
	{
		return;
	}
	SpeakerOutByte = Audio.spVol ? SpeakerOut
								 : (void (*)(uns8))audioDummyOutByte;
	AyOutByte = Audio.spVol ? AyOut
							: (void (*)(uns8, uns8))audioDummyOutByte;
}
unsigned int
AudioGetVolume(int channel)
{
	unsigned int volume;
	switch (channel)
	{
		case AUV_SP:
			volume = Audio.spVol;
			break;
		case AUV_AY:
			volume = Audio.ayVol;
			break;
		default:
			volume = 0;
	}
	return volume;
}
void
AudioPlayBuffer(void)
{
	if (Audio.active && !CFGGET(fastmode))
	{
		unsigned char *bpSp, *bpAyA, *bpAyB, *bpAyC, *bpOut;
		unsigned int volSp, volAy;
		int i;
		if ((volSp = Audio.spVol))
		{
			SpeakerFillBuffer();
		}
		if ((volAy = Audio.ayVol))
		{
			AyFillBuffer();
		}
		if ((volSp || volAy) != 0)
		{
			i = Audio.samplesPerFrame;
			bpOut = mixerBuffer;
			bpSp = Audio.spBuf;
			bpAyA = Audio.ayBuf;
			bpAyB = bpAyA + Audio.samplesPerFrame;
			bpAyC = bpAyB + Audio.samplesPerFrame;
			if (!Audio.stereo)
			{
				unsigned int val;
				while (i--)
				{
					val = (unsigned int)*bpSp++ * volSp +
						  (unsigned int)*bpAyA++ * volAy +
						  (unsigned int)*bpAyB++ * volAy +
						  (unsigned int)*bpAyC++ * volAy;
					*bpOut++ = (unsigned char)(val >> 3);
				}
			}
#ifdef REGISTERED
			else if (CFGGET(audioStereo) == AY_ACB)
			{
				unsigned int valL, valR;
				while (i--)
				{
					valL = valR = (unsigned int)*bpSp * volSp +
								  (unsigned int)*bpAyC * volAy;
					valL += (unsigned int)*bpAyA * volAy;
					valR += (unsigned int)*bpAyB * volAy;
					*bpOut++ = (unsigned char)(valL >> 3);
					*bpOut++ = (unsigned char)(valR >> 3);
					bpSp++;
					bpAyA++;
					bpAyB++;
					bpAyC++;
				}
			}
			else if (CFGGET(audioStereo) == AY_ABC)
			{
				unsigned int valL, valR;
				while (i--)
				{
					valL = valR = (unsigned int)*bpSp * volSp +
								  (unsigned int)*bpAyB * volAy;
					valL += (unsigned int)*bpAyA * volAy;
					valR += (unsigned int)*bpAyC * volAy;
					*bpOut++ = (unsigned char)(valL >> 3);
					*bpOut++ = (unsigned char)(valR >> 3);
					bpSp++;
					bpAyA++;
					bpAyB++;
					bpAyC++;
				}
			}
#endif
			else
			{
				unsigned int val;
				while (i--)
				{
					val = (unsigned int)*bpSp++ * volSp +
						  (unsigned int)*bpAyA++ * volAy +
						  (unsigned int)*bpAyB++ * volAy +
						  (unsigned int)*bpAyC++ * volAy;
					*bpOut++ = (unsigned char)(val >> 3);
					*bpOut++ = (unsigned char)(val >> 3);
				}
			}
			if (OsdPlayAudioBuffer(mixerBuffer, mixerSize) != mixerSize)
			{
				Msg(M_ERR, "couldn't write all bytes to audio device");
			}
		}
	}
	SpeakerResetBuffer();
	AyResetBuffer();
}
static void
audioDummyOutByte(void)
{
}
#endif
